# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections.abc import Sequence

import pytest

from astrapy import Collection, Database
from astrapy.constants import CallerType
from astrapy.settings.defaults import RERANKING_HEADER_API_KEY
from astrapy.utils.api_options import APIOptions, FullAPIOptions, defaultAPIOptions
from astrapy.utils.unset import _UNSET, UnsetType

from ..conftest import (
    DataAPICredentialsInfo,
    DefaultCollection,
)


def _wrapSomeOptions(
    src_database: Database,
    *,
    callers: Sequence[CallerType] | UnsetType = _UNSET,
    embedding_api_key: str | UnsetType = _UNSET,
    reranking_api_key: str | UnsetType = _UNSET,
) -> FullAPIOptions:
    return defaultAPIOptions(
        environment=src_database.api_options.environment
    ).with_override(
        APIOptions(
            callers=callers,
            embedding_api_key=embedding_api_key,
            reranking_api_key=reranking_api_key,
        )
    )


class TestCollectionsSync:
    @pytest.mark.describe("test of instantiating Collection, sync")
    def test_instantiate_collection_sync(
        self,
        sync_database: Database,
    ) -> None:
        col1: DefaultCollection = Collection(
            database=sync_database,
            name="id_test_collection",
            keyspace=None,
            api_options=_wrapSomeOptions(sync_database, callers=[("cn", "cv")]),
        )
        col2: DefaultCollection = Collection(
            database=sync_database,
            name="id_test_collection",
            keyspace=None,
            api_options=_wrapSomeOptions(sync_database, callers=[("cn", "cv")]),
        )
        assert col1 == col2

    @pytest.mark.describe("test of Collection conversions, sync")
    def test_convert_collection_sync(
        self,
        sync_database: Database,
    ) -> None:
        col1: DefaultCollection = Collection(
            database=sync_database,
            name="id_test_collection",
            keyspace=None,
            api_options=_wrapSomeOptions(sync_database, callers=[("cn", "cv")]),
        )
        assert col1 == col1._copy()
        assert col1 == col1.with_options()
        assert col1 == col1.to_async().to_sync()

    @pytest.mark.describe("test of Collection rich _copy, sync")
    def test_rich_copy_collection_sync(
        self,
        sync_database: Database,
    ) -> None:
        callers0 = [("cn", "cv"), ("dn", "dv")]
        col1: DefaultCollection = Collection(
            database=sync_database,
            name="id_test_collection",
            keyspace=None,
            api_options=_wrapSomeOptions(
                sync_database,
                callers=callers0,
                embedding_api_key="eak",
                reranking_api_key="rak",
            ),
        )

        col2 = col1._copy(
            embedding_api_key="zak",
        )
        assert col2 != col1
        col2b = col1._copy(
            reranking_api_key="zak",
        )
        assert col2b != col1

        assert col1.with_options(embedding_api_key="zak") != col1
        assert col1.with_options(reranking_api_key="zak") != col1

        assert (
            col1.with_options(embedding_api_key="zak").with_options(
                embedding_api_key="eak"
            )
            == col1
        )
        assert (
            col1.with_options(reranking_api_key="zak").with_options(
                reranking_api_key="rak"
            )
            == col1
        )

    @pytest.mark.describe("test of Collection rich conversions, sync")
    def test_rich_convert_collection_sync(
        self,
        sync_database: Database,
    ) -> None:
        callers0 = [("cn", "cv"), ("dn", "dv")]
        col1: DefaultCollection = Collection(
            database=sync_database,
            name="id_test_collection",
            keyspace="the_ks",
            api_options=_wrapSomeOptions(
                sync_database,
                callers=callers0,
                embedding_api_key="eak",
                reranking_api_key="rak",
            ),
        )
        assert col1 != col1.to_async(embedding_api_key="zak").to_sync()
        assert col1 != col1.to_async(reranking_api_key="zak").to_sync()

        col2a = col1.to_async(
            embedding_api_key="zak",
        )
        assert col2a.to_sync() != col1
        col2ba = col1.to_async(
            reranking_api_key="zak",
        )
        assert col2ba.to_sync() != col1

        col3 = col2a.to_sync(
            embedding_api_key="eak",
        )
        assert col3 == col1
        col3b = col2ba.to_sync(
            reranking_api_key="rak",
        )
        assert col3b == col1

    @pytest.mark.describe("test of Collection database property, sync")
    def test_collection_database_property_sync(
        self,
    ) -> None:
        opts0 = defaultAPIOptions(environment="other")
        db1 = Database(api_endpoint="a", keyspace="ns1", api_options=opts0)
        db2 = Database(api_endpoint="a", keyspace="ns2", api_options=opts0)
        col1: DefaultCollection = Collection(
            database=db1,
            name="coll",
            keyspace=None,
            api_options=opts0,
        )
        col2: DefaultCollection = Collection(
            database=db1, name="coll", api_options=opts0, keyspace="ns2"
        )
        assert col1.database == db1
        assert col2.database == db2

    @pytest.mark.describe("test of Collection name property, sync")
    def test_collection_name_property_sync(
        self,
    ) -> None:
        opts0 = defaultAPIOptions(environment="other")
        db1 = Database(api_endpoint="a", keyspace="ns1", api_options=opts0)
        col1: DefaultCollection = Collection(
            database=db1,
            name="coll",
            keyspace=None,
            api_options=opts0,
        )
        assert col1.name == "coll"

    @pytest.mark.describe("test collection keyspace property, sync")
    def test_collection_keyspace_sync(
        self,
        sync_database: Database,
        data_api_credentials_info: DataAPICredentialsInfo,
    ) -> None:
        col1 = sync_database.get_collection("id_test_collection")
        assert col1.keyspace == sync_database.keyspace

        col2 = sync_database.get_collection(
            "id_test_collection",
            keyspace=data_api_credentials_info["secondary_keyspace"],
        )
        assert col2.keyspace == data_api_credentials_info["secondary_keyspace"]

        col3: DefaultCollection = Collection(
            database=sync_database,
            name="id_test_collection",
            keyspace=None,
            api_options=sync_database.api_options,
        )
        assert col3.keyspace == sync_database.keyspace

        col4: DefaultCollection = Collection(
            database=sync_database,
            name="id_test_collection",
            keyspace=data_api_credentials_info["secondary_keyspace"],
            api_options=sync_database.api_options,
        )
        assert col4.keyspace == data_api_credentials_info["secondary_keyspace"]
        assert col1 == col3
        assert col2 == col4

    @pytest.mark.describe(
        "test collection reranking API key in commander headers, sync"
    )
    def test_collection_rerankingapikey_in_headers_sync(
        self,
        sync_database: Database,
    ) -> None:
        col_0 = sync_database.get_collection("q")
        col_1 = sync_database.get_collection("q", reranking_api_key="RAK")
        assert RERANKING_HEADER_API_KEY not in col_0._api_commander.full_headers
        assert col_1._api_commander.full_headers[RERANKING_HEADER_API_KEY] == "RAK"



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from dataclasses import dataclass
from typing import Generic

from astrapy.data.cursors.cursor import TRAW
from astrapy.data_types import DataAPIVector


@dataclass
class FindPage(Generic[TRAW]):
    """
    A whole pageful of results from a find operation. This object represents
    the form taken by the returned items when using the pagination interface
    explicitly.

    Attributes:
        results: the list of entries obtained on the retrieved page (possibly
            after applying a mapping function, if one is specified in the cursor).
        next_page_state: a string encoding the pagination state. If the find
            operation does not admit any further page, this is returned as None.
            Otherwise, its value can be used to resume consuming the `find`
            results on another cursor instantiated independently later on.
        sort_vector: if the find operation was done with the "include
            sort vector" flag set to True, and the sort criterion is a vector sorting,
            this contains the query vector used for the search. The query vector is
            expressed as a list of floats or a DataAPIVector depending on the serdes
            settings for the collection/table that originated the cursor.
            If not applicable, this attribute is returned as None.
    """

    results: list[TRAW]
    next_page_state: str | None
    sort_vector: list[float] | DataAPIVector | None

    def __repr__(self) -> str:
        pieces = [
            pc
            for pc in (
                f"results=<{len(self.results)} entries>",
                "next_page_state=..." if self.next_page_state else None,
                "sort_vector=..." if self.sort_vector else None,
            )
            if pc is not None
        ]
        return f"{self.__class__.__name__}({', '.join(pieces)})"


@dataclass
class FindAndRerankPage(Generic[TRAW]):
    """
    A whole pageful of results from a findAndRerank operation. This object represents
    the form taken by the returned items when using the pagination interface
    explicitly.

    Attributes:
        results: the list of entries obtained on the retrieved page (possibly
            after applying a mapping function, if one is specified in the cursor).
            In absence of mapping functions, this is a list of RerankedResult objects.
        next_page_state: a string encoding the pagination state. If the find-and-rerank
            operation does not admit any further page, this is returned as None.
            Otherwise, its value can be used to resume consuming the `find_and_rerank`
            results on another cursor instantiated independently later on.
        sort_vector: if the find-and-rerank operation was done with the "include
            sort vector" flag set to True, and the sort criterion is a vector sorting,
            this contains the query vector used for the search. The query vector is
            expressed as a list of floats or a DataAPIVector depending on the serdes
            settings for the collection/table that originated the cursor.
            If not applicable, this attribute is returned as None.
    """

    results: list[TRAW]
    next_page_state: str | None
    sort_vector: list[float] | DataAPIVector | None

    def __repr__(self) -> str:
        pieces = [
            pc
            for pc in (
                f"results=<{len(self.results)} entries>",
                "next_page_state=..." if self.next_page_state else None,
                "sort_vector=..." if self.sort_vector else None,
            )
            if pc is not None
        ]
        return f"{self.__class__.__name__}({', '.join(pieces)})"



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections.abc import Iterable, Iterator
from typing import Protocol, TypeVar

# TODO: once supporting 3.11+, this can be simplified.
# 'Self' is imported differently below 3.11, cf. below. Then:
# 1. To appease typecheckers on e.g. 3.9, we need the `attr-defined` exception
# 2. To appease typecheckers on 3.11+, we need the `unused-ignore` exception.
try:
    # Python 3.11 upward:
    from typing import Self  # type:ignore[attr-defined,unused-ignore]
except ImportError:
    # 3.10 and below:
    from typing_extensions import Self

from contextlib import contextmanager

from uuid6 import uuid7

from astrapy.api_options import APIOptions
from astrapy.event_observers.events import ObservableEvent, ObservableEventType
from astrapy.event_observers.observers import Observer
from astrapy.utils.unset import _UNSET, UnsetType


class OptionAwareDatabaseObject(Protocol):
    def with_options(
        self,
        *,
        api_options: APIOptions | UnsetType = _UNSET,
    ) -> Self: ...


DB_OBJ = TypeVar("DB_OBJ", bound=OptionAwareDatabaseObject)


@contextmanager
def event_collector(
    target: DB_OBJ,
    *,
    destination: list[ObservableEvent]
    | dict[ObservableEventType, list[ObservableEvent]],
    event_types: Iterable[ObservableEventType] | None = None,
) -> Iterator[DB_OBJ]:
    """
    Create a context manager wrapping a list (of `ObservableEvent` objects) or
    a dict (with `ObservableEventType` keys and `list[ObservableEvent]` values),
    for quickly instrumenting a client, database, table, collection or admin object.

    This utility is meant to be used in a `with` statement, so that events emitted
    by the instrumented classes within the with block are captured into the provided
    lists/dictionaries. Events emitted by spawned classes (e.g. when a Database
    creates a Collection in the `with` block) are also received this way.

    Once outside the `with` block, collection of events stops, but the provided
    destination can still be accessed according to its ordinary scoping.

    Args:
        target: an object that issues Data API / DevOps API requests. The target
            must have a `with_options` suitable method: meaning, it can be any of
            the following: `DataAPIClient`, `AsyncDatabase`, `Database`,
            `AsyncCollection`, `Collection`, `AsyncTable`, `Table`, `AstraDBAdmin`,
            `AstraDBDatabaseAdmin`, `DataAPIDatabaseAdmin`.
        destination: a list or a dictionary where the collected events will be stored.
            For dictionaries, events are grouped into lists, one per each event type,
            stored under the corresponding `ObservableEventType` value as dict key.
        event_types: if provided, it's a list of event types so that only
            events matching this filter are processed.

    Returns:
        Yields an instrumented version of the input target, with an added observer
        set to accumulate the received events into the provided destination. Any
        pre-existing observer is untouched.

    Example:
        >>> ev_lst: list[ObservableEvent] = []
        >>> with event_collector(db, destination=ev_lst) as instrumented_db:
        ...     _ = instrumented_db.list_table_names()
        ...     table = instrumented_db.get_table("my_table")
        ...     _ = table.find_one({"k": 101})
        ...
        >>> print(len(ev_lst))
        5
        >>> print(ev_lst[0].event_type)
        ObservableEventType.REQUEST
    """
    observer_id_ = f"observer_{str(uuid7())}"
    observer_: Observer
    if isinstance(destination, list):
        observer_ = Observer.from_event_list(destination, event_types=event_types)
    else:
        observer_ = Observer.from_event_dict(destination, event_types=event_types)
    api_options = APIOptions(event_observers={observer_id_: observer_})
    target_ = target.with_options(api_options=api_options)
    try:
        yield target_
    finally:
        observer_.enabled = False



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from astrapy.data.info.table_descriptor.table_columns import TableAPISupportDescriptor
from astrapy.data.info.table_descriptor.type_creation import CreateTypeDefinition
from astrapy.data.utils.table_types import (
    TableUDTColumnType,
    TableUnsupportedColumnType,
)
from astrapy.utils.parsing import _warn_residual_keys


@dataclass
class ListTypeDescriptor:
    """
    A structure describing a user-defined type (UDT) stored on the database.

    This object is used for the items returned by the database `list_types` method.
    `ListTypeDescriptor` expresses all information received by the Data API, including
    (when provided) the UDT name as found on the database, the UDT name and possibly a
    sub-object detailing the allowed operations with the UDT.

    This object must be able to describe any item returned from the Data API:
    this means it can describe "unsupported" UDTs as well (i.e. those which have been
    created outside of the Data API). Unsupported UDTs lack some attributes compared
    to the fully-supported ones.

    Attributes:
        udt_type: a value of either the TableUDTColumnType or
            the TableUnsupportedColumnType enum, depending on the UDT support status.
        udt_name: the name of the UDT as is stored in the database (and in a keyspace).
        definition: the definition of the type, i.e. its fields and their types.
        api_support: a structure detailing what operations the type supports.
    """

    udt_type: TableUDTColumnType | TableUnsupportedColumnType
    udt_name: str | None
    definition: CreateTypeDefinition | None
    api_support: TableAPISupportDescriptor | None

    def __init__(
        self,
        *,
        udt_type: TableUDTColumnType | TableUnsupportedColumnType,
        udt_name: str | None,
        definition: CreateTypeDefinition | None,
        api_support: TableAPISupportDescriptor | None,
    ) -> None:
        self.udt_type = udt_type
        self.udt_name = udt_name
        self.definition = definition
        self.api_support = api_support

    def __repr__(self) -> str:
        if isinstance(self.udt_type, TableUnsupportedColumnType):
            return f"{self.__class__.__name__}({self.udt_type.value})"
        else:
            return f"{self.__class__.__name__}({self.udt_name}: {self.definition})"

    @staticmethod
    def _is_valid_dict(raw_dict: dict[str, Any]) -> bool:
        """
        Assess whether a dictionary can be converted into a ListTypeDescriptor.

        This can be used by e.g. the database `list_types` method to filter
        offending responses and issue warnings if needed.

        Returns:
            True if and only if the dict is valid, otherwise False.
        """

        return all(fld in raw_dict for fld in {"type", "apiSupport"})

    def as_dict(self) -> dict[str, Any]:
        """Recast this object into a dictionary."""

        return {
            k: v
            for k, v in {
                "type": self.udt_type.value,
                "udtName": self.udt_name,
                "definition": self.definition.as_dict()
                if self.definition is not None
                else None,
                "apiSupport": self.api_support.as_dict()
                if self.api_support is not None
                else None,
            }.items()
            if v is not None
        }

    @classmethod
    def _from_dict(cls, raw_dict: dict[str, Any]) -> ListTypeDescriptor:
        """
        Create an instance of ListTypeDescriptor from a dictionary
        such as one from the Data API.
        """

        _warn_residual_keys(
            cls, raw_dict, {"type", "udtName", "definition", "apiSupport"}
        )
        _udt_type: TableUDTColumnType | TableUnsupportedColumnType
        if raw_dict["type"] in TableUDTColumnType:
            _udt_type = TableUDTColumnType.coerce(raw_dict["type"])
        else:
            _udt_type = TableUnsupportedColumnType.coerce(raw_dict["type"])
        return ListTypeDescriptor(
            udt_type=_udt_type,
            udt_name=raw_dict.get("udtName"),
            definition=CreateTypeDefinition._from_dict(raw_dict["definition"])
            if "definition" in raw_dict
            else None,
            api_support=TableAPISupportDescriptor._from_dict(raw_dict["apiSupport"])
            if "apiSupport" in raw_dict
            else None,
        )

    @classmethod
    def coerce(
        cls, raw_input: ListTypeDescriptor | dict[str, Any]
    ) -> ListTypeDescriptor:
        """
        Normalize the input, whether an object already or a plain dictionary
        of the right structure, into a ListTypeDescriptor.
        """

        if isinstance(raw_input, ListTypeDescriptor):
            return raw_input
        else:
            return cls._from_dict(raw_input)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy.utils.document_paths import (
    ILLEGAL_ESCAPE_ERROR_MESSAGE_TEMPLATE,
    UNTERMINATED_ESCAPE_ERROR_MESSAGE_TEMPLATE,
    escape_field_names,
    unescape_field_path,
)

ESCAPE_TEST_FIELDS = {
    "a": "a",
    "a.b": "a&.b",
    "a&c": "a&&c",
    "a..": "a&.&.",
    "": "",
    "xyz": "xyz",
    ".": "&.",
    "...": "&.&.&.",
    "&": "&&",
    "&&": "&&&&",
    "&.&.&..": "&&&.&&&.&&&.&.",
    "🇦🇨": "🇦🇨",
    "q🇦🇨w": "q🇦🇨w",
    ".🇦🇨&": "&.🇦🇨&&",
    "a&b.c.&.d": "a&&b&.c&.&&&.d",
    ".env": "&.env",
    "&quot": "&&quot",
    "ݤ": "ݤ",
    "Aݤ": "Aݤ",
    "ݤZ": "ݤZ",
    "🚾 🆒 🆓 🆕 🆖 🆗 🆙 🏧": "🚾 🆒 🆓 🆕 🆖 🆗 🆙 🏧",
    "✋🏿 💪🏿 👐🏿 🙌🏿 👏🏿 🙏🏿": "✋🏿 💪🏿 👐🏿 🙌🏿 👏🏿 🙏🏿",
    "👨‍👩‍👦 👨‍👩‍👧‍👦 👨‍👨‍👦 👩‍👩‍👧 👨‍👦 👨‍👧‍👦 👩‍👦 👩‍👧‍👦": "👨‍👩‍👦 👨‍👩‍👧‍👦 👨‍👨‍👦 👩‍👩‍👧 👨‍👦 👨‍👧‍👦 👩‍👦 👩‍👧‍👦",
}


class TestDocumentPaths:
    @pytest.mark.describe("test of escape_field_names, one arg")
    def test_escape_field_names_onearg(self) -> None:
        for lit_fn, esc_fn in ESCAPE_TEST_FIELDS.items():
            assert escape_field_names(lit_fn) == esc_fn
            assert escape_field_names([lit_fn]) == esc_fn
        for num in [0, 12, 130099]:
            assert escape_field_names(num) == str(num)
            assert escape_field_names([num]) == str(num)

    @pytest.mark.describe("test of escape_field_names, multiple args")
    def test_escape_field_names_multiargs(self) -> None:
        all_lits, all_escs = list(zip(*ESCAPE_TEST_FIELDS.items()))
        assert escape_field_names(all_lits) == ".".join(all_escs)
        assert escape_field_names(*all_lits) == ".".join(all_escs)
        assert escape_field_names(all_lits[:0]) == ".".join(all_escs[:0])
        assert escape_field_names(*all_lits[:0]) == ".".join(all_escs[:0])
        assert escape_field_names(all_lits[:3]) == ".".join(all_escs[:3])
        assert escape_field_names(*all_lits[:3]) == ".".join(all_escs[:3])
        assert escape_field_names(all_lits[3:6]) == ".".join(all_escs[3:6])
        assert escape_field_names(*all_lits[3:6]) == ".".join(all_escs[3:6])

        assert escape_field_names(["first", 12, "last&!."]) == "first.12.last&&!&."
        assert escape_field_names("first", 12, "last&!.") == "first.12.last&&!&."

    @pytest.mark.describe("test of unescape_field_path")
    def test_unescape_field_path(self) -> None:
        assert unescape_field_path("a&.b") == ["a.b"]

        all_lits, all_escs = list(zip(*ESCAPE_TEST_FIELDS.items()))
        assert unescape_field_path(".".join(all_escs)) == list(all_lits)
        assert unescape_field_path(".".join(all_escs[:3])) == list(all_lits[:3])
        assert unescape_field_path(".".join(all_escs[3:6])) == list(all_lits[3:6])

        assert unescape_field_path("") == []

        with pytest.raises(
            ValueError, match=ILLEGAL_ESCAPE_ERROR_MESSAGE_TEMPLATE[:24]
        ):
            unescape_field_path("a.b&?c.d")

        with pytest.raises(
            ValueError, match=UNTERMINATED_ESCAPE_ERROR_MESSAGE_TEMPLATE[:24]
        ):
            unescape_field_path("a.b.c&")
